import { CliRpcClient } from '../../util/rpc-client';
import { CliAuthenticationClient } from '../../util/authentication-client';
import { AppJsonFile } from '../../util/app-json-file';
import { Spinner } from '../../util/spinner';
import { getSystemId } from '../../util/system-id';
import inquirer = require('inquirer');
import { JsSpawner } from '../../util/spawner/js-spawner';
import { SystemDomainName } from '@alwaysai/cloud-api';
import fetch from 'node-fetch';
import { cd } from '../../util/cd';
import { machineId, machineIdSync } from 'node-machine-id';
const os = require('os');
import { TargetJsonFile } from '../../util/target-json-file';

const util = require('util');
const fs = require('fs');
const streamPipeline = util.promisify(require('stream').pipeline);

import { CliTerseError } from '@alwaysai/alwayscli';
import { join } from 'path';
import rimraf = require('rimraf');
import mkdirp = require('mkdirp');
import { PLEASE_REPORT_THIS_ERROR_MESSAGE, ALWAYSAI_STARTER_APPS } from '../../constants';
import { renameSync, readdirSync, existsSync } from 'fs';
import * as tar from 'tar';
import pump = require('pump');

import { appConfigurePreliminaryStepsComponent } from '../app-configure-preliminary-steps-component';

async function checkForProjectId() {
  const username = await CliAuthenticationClient().getInfo();

  const listProjectsUserIsCollaborator = await CliRpcClient().listProjectsUserIsCollaborator(
    {
      user_name: username.username,
    },
  );

  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  if (project) {
    const collabProject = listProjectsUserIsCollaborator.filter(value => {
      if (project.projectId === value.id.toString()) {
        return value.id.toString();
      }
    });
    if (!collabProject.length) {
      return false;
    }
    return collabProject;
  }
}

async function selectFromListOfProjects() {
  const username = await CliAuthenticationClient().getInfo();

  const listProjectsUserIsCollaborator = await CliRpcClient().listProjectsUserIsCollaborator(
    {
      user_name: username.username,
    },
  );

  // check if listNames is not empty

  const questions = [
    {
      type: 'list',
      name: 'projects',
      message:
        'Select from list of projects or create a new project at (https://dashboard.a6i1.net/dashboard/create-team):',
      choices: listProjectsUserIsCollaborator,
      filter(val) {
        return val.toLowerCase();
      },
    },
  ];

  return await inquirer.prompt(questions).then(answer => {
    const pickedProject = listProjectsUserIsCollaborator.filter(value => {
      if (value.name.toLocaleLowerCase() === answer.projects) {
        return value;
      }
    });

    return pickedProject[0];
  });
}

async function getListOfProjects() {
  const username = await CliAuthenticationClient().getInfo();

  const listProjectsUserIsCollaborator = await CliRpcClient().listProjectsUserIsCollaborator(
    {
      user_name: username.username,
    },
  );
  return listProjectsUserIsCollaborator;
}

async function selectTypeOfProject() {
  const questions = [
    {
      type: 'list',
      name: 'projects',
      message: 'Select Project source',
      choices: ['Use starter app', 'Start Blank'],
    },
  ];

  return await inquirer.prompt(questions).then(answer => {
    return answer;
  });
}

async function localProject() {
  // check if project has a git url`
  const spawner = JsSpawner();

  spawner.runForegroundSync({
    exe: 'git',
    args: ['clone', 'https://github.com/alwaysai/alwaysai-twilio-sms'],
  });
}

async function readJson() {
  const systemDomainName = SystemDomainName(getSystemId());
  const starterAppsUrl = `https://dashboard.${systemDomainName}/docs/apps-test-json.json`;
  const authorizationHeader = await CliAuthenticationClient().getAuthorizationHeader();
  const jwt = authorizationHeader.Authorization.split(' ')[1];
  const cookieHeader = {
    cookie: `token=${jwt}`,
  };
  try {
    const response = await fetch(starterAppsUrl, {
      redirect: 'manual',
      headers: { ...cookieHeader },
    });
    const readerStream = await response.buffer();
    const parsedJSon = JSON.parse(readerStream.toString());

    const questions = [
      {
        type: 'list',
        name: 'starterApps',
        message: 'Select a starter app):',
        choices: parsedJSon['starter-apps'],
        filter(val) {
          return val.toLowerCase();
        },
      },
    ];

    return await inquirer.prompt(questions).then(answer => {
      return answer['starterApps'];
    });
  } finally {
  }
}

async function downloadSingleApp(dir = process.cwd(), pickedStarterApp) {
  const systemDomainName = SystemDomainName(getSystemId());

  const starterAppsUrl = `https://dashboard.${systemDomainName}/docs/${pickedStarterApp}.tar.gz`;

  const destinationDir = join(dir, pickedStarterApp);

  if (existsSync(destinationDir)) {
    throw new CliTerseError(`Directory "${destinationDir}" already exists`);
  }
  const tmpDir = `${destinationDir}.download`;

  rimraf.sync(tmpDir);
  const authorizationHeader = await CliAuthenticationClient().getAuthorizationHeader();
  const jwt = authorizationHeader.Authorization.split(' ')[1];
  const cookieHeader = {
    cookie: `token=${jwt}`,
  };
  mkdirp.sync(tmpDir);
  try {
    const response = await fetch(starterAppsUrl, {
      redirect: 'manual',
      headers: { ...cookieHeader },
    });
    if (!response.ok) {
      throw new CliTerseError(
        `Server responded ${response.status} "${response.statusText}"`,
      );
    }

    await new Promise((resolve, reject) => {
      pump([response.body, tar.extract({ cwd: tmpDir })], err => {
        if (err) {
          reject(err);
        } else {
          resolve();
        }
      });
    });
    const fileNames = readdirSync(tmpDir);

    if (fileNames.length !== 1 || !fileNames[0]) {
      throw new Error(
        `Expected ${ALWAYSAI_STARTER_APPS} package to contain a single directory. ${PLEASE_REPORT_THIS_ERROR_MESSAGE}`,
      );
    }

    await renameSync(join(tmpDir, fileNames[0]), destinationDir);

    const starterAppFiles = readdirSync(destinationDir);

    starterAppFiles.forEach(file => {
      renameSync(join(destinationDir, file), join(dir, file));
    });
  } finally {
    rimraf.sync(tmpDir);
    rimraf.sync(destinationDir);
  }
}

async function newConfig() {
  const appJsonFile = AppJsonFile();

  if (appJsonFile.exists()) {
    await checkForProjectId();
  } else {
    const selectedProject = await selectFromListOfProjects();
    // check if it has a git url
    if (selectedProject.git_url) {
      console.log('this is the git url process');
      const spawner = JsSpawner();

      spawner.runForegroundSync({
        exe: 'git',
        args: ['clone', selectedProject.git_url],
      });
    } else {
      const projectSource = await selectTypeOfProject();

      if (projectSource.projects === 'Use starter app') {
        const pickedStarterApp = await readJson();
        await downloadSingleApp(process.cwd(), pickedStarterApp);

        cd(`${process.cwd()}/${pickedStarterApp}`);
        const newJson = AppJsonFile();

        newJson.addProjectId(selectedProject['id'].toString());
      } else {
        const yes = false;
        await appConfigurePreliminaryStepsComponent({ yes });
        console.log('create blank project');
        appJsonFile.addProjectId(selectedProject['id'].toString());
      }
    }
  }
}

async function configFlow() {
  const appJsonFile = AppJsonFile();

  if (appJsonFile.exists()) {
    const selectedProject = await selectFromListOfProjects();
    // check if it has a git url
    if (selectedProject.git_url) {
      console.log('this is the git url process');
      const spawner = JsSpawner();

      spawner.runForegroundSync({
        exe: 'git',
        args: ['clone', selectedProject.git_url],
      });
    } else {
      const projectSource = await selectTypeOfProject();

      if (projectSource.projects === 'Use starter app') {
        const pickedStarterApp = await readJson();
        await downloadSingleApp(process.cwd(), pickedStarterApp);

        cd(`${process.cwd()}/${pickedStarterApp}`);
        const newJson = AppJsonFile();

        newJson.addProjectId(selectedProject['id'].toString());
      } else {
        const yes = false;
        console.log('create blank project');
        appJsonFile.addProjectId(selectedProject['id'].toString());
      }
    }
  }
}

async function initWithArgs() {
  const appJsonFile = AppJsonFile();

  if (appJsonFile.exists()) {
    await checkForProjectId();
  } else {
    const selectedProject = await selectFromListOfProjects();
    // check if it has a git url
    if (selectedProject.git_url) {
      console.log('this is the git url process');
      const spawner = JsSpawner();

      spawner.runForegroundSync({
        exe: 'git',
        args: ['clone', selectedProject.git_url],
      });
    } else {
      const projectSource = await selectTypeOfProject();

      if (projectSource.projects === 'Use starter app') {
        const pickedStarterApp = await readJson();
        await downloadSingleApp(process.cwd(), pickedStarterApp);

        cd(`${process.cwd()}/${pickedStarterApp}`);
        const newJson = AppJsonFile();

        newJson.addProjectId(selectedProject['id'].toString());
      } else {
        const yes = false;
        await appConfigurePreliminaryStepsComponent({ yes });
        console.log('create blank project');
        appJsonFile.addProjectId(selectedProject['id'].toString());
      }
    }
  }
}

async function deviceFriendlyName() {
  const questions = [
    {
      type: 'input',
      name: 'deviceFriendlyName',
      message: 'Enter a device friendly name:',
    },
  ];

  return await inquirer.prompt(questions).then(answer => {
    console.log({ answer });
    return answer.deviceFriendlyName;
  });
}

async function checkIfDeviceHasBeenRegistered() {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;

  if (project) {
    const projectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    const machineId = machineIdSync();

    const newProjectDevices = projectDevices.filter(value => {
      if (value.hardware_ids === machineId) {
        return true;
      }
    });

    if (newProjectDevices[0]) {
      return true;
    }
  }

  return false;
}

async function sshCheckIfDeviceHasBeenRegistered(machineId) {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;

  if (project) {
    const projectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    const newProjectDevices = projectDevices.filter(value => {
      if (value.hardware_ids === machineId) {
        return true;
      }
    });

    if (newProjectDevices[0]) {
      return true;
    }
  }

  return false;
}

async function storeDeviceId() {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;

  const checkDeviceLimit = await checkTeamLimit();

  if (checkDeviceLimit) {
    const inputDeviceFriendlyName = await deviceFriendlyName();

    const machineId = machineIdSync();
    const username = await CliAuthenticationClient().getInfo();
    const hostname = os.hostname();
    const deviceUserName = os.userInfo().username;
    // check if device has already been registered
    const isDeviceRegistered = await checkIfDeviceHasBeenRegistered();
    if (!isDeviceRegistered) {
      const createDevice = await CliRpcClient().createDevice({
        owner: username.username,
        friendly_name: inputDeviceFriendlyName,
        host_name: hostname,
        device_user_name: deviceUserName,
        description: '',
        hardware_ids: machineId,
        device_hash: '',
      });
      const appJsonFile = AppJsonFile();
      const project = appJsonFile.read().project;
      if (project) {
        const addDeviceToProject = await CliRpcClient().addProjectDevice({
          device_id: createDevice.id,
          project_id: Number(project.projectId),
        });

        await saveDeviceJson(createDevice.id.toString());
      }
    } else {
      throw new CliTerseError(
        'You already registered this device please pick a saved device',
      );
    }
  } else {
    throw new CliTerseError('You are at your limit');
  }
}

async function sshStoreDeviceId(machineId) {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  const targetJsonFile = TargetJsonFile();

  const targetHostSpawner = targetJsonFile.readHostSpawner();

  const checkDeviceLimit = await checkTeamLimit();

  const hostname = await targetHostSpawner.run({
    exe: 'hostname',
  });

  const deviceUserName = await targetHostSpawner.run({
    exe: 'whoami',
  });

  if (checkDeviceLimit) {
    const inputDeviceFriendlyName = await deviceFriendlyName();

    const username = await CliAuthenticationClient().getInfo();

    // check if device has already been registered
    const isDeviceRegistered = await sshCheckIfDeviceHasBeenRegistered(machineId);
    if (!isDeviceRegistered) {
      const createDevice = await CliRpcClient().createDevice({
        owner: username.username,
        friendly_name: inputDeviceFriendlyName,
        host_name: hostname,
        device_user_name: deviceUserName,
        description: '',
        hardware_ids: machineId,
        device_hash: '',
      });
      const appJsonFile = AppJsonFile();
      const project = appJsonFile.read().project;
      if (project) {
        const addDeviceToProject = await CliRpcClient().addProjectDevice({
          device_id: createDevice.id,
          project_id: Number(project.projectId),
        });

        await saveDeviceJson(createDevice.id.toString());
      }
    } else {
      throw new CliTerseError(
        'You already registered this device please pick a saved device in this remote device',
      );
    }
  } else {
    throw new CliTerseError('You are at your limit');
  }
}

async function checkTeamLimit() {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  if (project) {
    const teamId = await CliRpcClient().getProject({ id: Number(project.projectId) });
    const getDeviceLimits = await CliRpcClient().checkTeamLimits({
      team_id: teamId.team_id,
    });
    const devicesLimit = getDeviceLimits['devices'].limit;
    const devicesUsed = getDeviceLimits['devices'].used;

    if (devicesLimit > devicesUsed) {
      return true;
    }
  }

  return false;
}

async function checkIfDeviceIsBeingUsed(hardwareIds: string) {
  const machineId = machineIdSync();

  if (machineId === hardwareIds) {
    console.log(true);
    return true;
  }
  return false;
}

async function sshCheckIfDeviceIsBeingUsed(hardwareIds: string, machineId: string) {
  console.log({ hardwareIds });
  console.log({ machineId });

  if (machineId === hardwareIds) {
    return true;
  }
  return false;
}

async function saveDeviceJson(id: string) {
  // saves device id in json
  const appJsonFile = AppJsonFile();

  appJsonFile.addDeviceId(id);
}

async function createNewOrPick() {
  const noSaved = await checkIfAnySavedDevices();

  let choices: any = [];

  if (noSaved) {
    choices = ['Use saved device', 'Create new device'];
  } else {
    choices = ['Create new device'];
  }

  const questions = [
    {
      type: 'list',
      name: 'deviceChoice',
      message: 'Select choice:',
      choices,
    },
  ];

  return await inquirer.prompt(questions).then(answer => {
    return answer.deviceChoice;
  });
}

async function sshCreateNewOrPick(machineId) {
  const noSaved = await sshCheckIfAnySavedDevices(machineId);

  let choices: any = [];

  if (noSaved) {
    choices = ['Use saved device', 'Create new device'];
  } else {
    choices = ['Create new device'];
  }

  const questions = [
    {
      type: 'list',
      name: 'deviceChoice',
      message: 'Select choice:',
      choices,
    },
  ];

  return await inquirer.prompt(questions).then(answer => {
    return answer.deviceChoice;
  });
}

async function checkIfAnySavedDevices() {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  if (project) {
    const projectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    if (projectDevices.length) {
      return true;
    }
  }
  return false;
}

async function sshCheckIfAnySavedDevices(machineId) {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  if (project) {
    const projectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    if (projectDevices.length) {
      return true;
    }
  }
  return false;
}

async function pickDevice() {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  if (project) {
    const projectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    const newObject: any = [];
    projectDevices.forEach(value => {
      if (value) {
        newObject.push({
          name: value.friendly_name,
          value: { id: value.id, hardwareIds: value.hardware_ids },
        });
      }
    });

    const questions = [
      {
        type: 'list',
        name: 'device',
        message: 'Select a device:',
        choices: newObject,
      },
    ];

    return await inquirer.prompt(questions).then(answer => {
      return answer.device;
    });
  }

  // check json file if there is a device id present
}

async function pickAndStoreDevice() {
  const pickedDevice = await pickDevice();
  // reference device to current device

  const canIUse = await checkIfDeviceIsBeingUsed(pickedDevice.hardwareIds);

  if (canIUse) {
    await saveDeviceJson(pickedDevice.id.toString());
  } else {
    console.log('pick another device');
  }
}

async function sshPickAndStoreDevice(machineId) {
  const pickedDevice = await pickDevice();
  // reference device to current device

  const canIUse = await sshCheckIfDeviceIsBeingUsed(pickedDevice.hardwareIds, machineId);

  if (canIUse) {
    await saveDeviceJson(pickedDevice.id.toString());
  } else {
    throw new CliTerseError(
      'The device picked does not match selected device. Please pick another one or make a new one.',
    );
  }
}

async function deviceFlow() {
  const firstCheck = await checkIfDeviceIsStored();

  if (!firstCheck) {
    const deviceSelection = await createNewOrPick();
    if (deviceSelection === 'Use saved device') {
      await pickAndStoreDevice();
    } else {
      await storeDeviceId();
    }
  }
}

async function sshDeviceFlow(machineId) {
  const firstCheck = await sshCheckIfDeviceIsStored(machineId);

  if (!firstCheck) {
    const deviceSelection = await sshCreateNewOrPick(machineId);
    if (deviceSelection === 'Use saved device') {
      await sshPickAndStoreDevice(machineId);
    } else {
      await sshStoreDeviceId(machineId);
    }
  }
}

async function checkIfDeviceIsStored() {
  const machineId = machineIdSync();

  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  const device = appJsonFile.read().device;

  if (project && device) {
    const getProjectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    // need to validate with json file also

    const validDevice = getProjectDevices.filter(value => {
      if (value.id === Number(device.deviceId)) {
        return value;
      }
    });

    if (validDevice[0]) {
      if (validDevice[0].hardware_ids === machineId) {
        return true;
      }
    }
  }
  return false;
}

async function sshCheckIfDeviceIsStored(machineId) {
  const appJsonFile = AppJsonFile();
  const project = appJsonFile.read().project;
  const device = appJsonFile.read().device;

  if (project && device) {
    const getProjectDevices = await CliRpcClient().getProjectDevices({
      project_id: Number(project.projectId),
    });

    // need to validate with json file also

    const validDevice = getProjectDevices.filter(value => {
      if (value.id === Number(device.deviceId)) {
        return value;
      }
    });

    if (validDevice[0]) {
      if (validDevice[0].hardware_ids === machineId) {
        return true;
      }
    }
  }
  return false;
}

async function checkIfUserBelongsToProject() {}
export {
  checkForProjectId,
  selectFromListOfProjects,
  localProject,
  readJson,
  downloadSingleApp,
  newConfig,
  initWithArgs,
  configFlow,
  selectTypeOfProject,
  checkIfUserBelongsToProject,
  getListOfProjects,
  storeDeviceId,
  checkTeamLimit,
  pickDevice,
  saveDeviceJson,
  checkIfDeviceIsBeingUsed,
  pickAndStoreDevice,
  createNewOrPick,
  deviceFlow,
  checkIfDeviceIsStored,
  checkIfDeviceHasBeenRegistered,
  checkIfAnySavedDevices,
  sshDeviceFlow,
};
