import { targetProtocolPromptComponent } from './target-protocol-prompt-component';
import { TargetProtocol } from '../util/target-protocol';
import { checkForDockerComponent } from './check-for-docker-component';
import { TargetJsonFile } from '../util/target-json-file';
import { targetHostnamePromptComponent } from './target-hostname-prompt-component';
import { targetPathPromptComponent } from './target-path-prompt-component';
import {
  DOCKER_IMAGE_ID_INITIAL_VALUE,
  SUPPORTED_JETPACK_VERSION,
  JETPACK_IMAGE_LINK,
} from '../constants';
import { writeTargetJsonFileComponent } from './write-target-json-file-component';
import { findOrWritePrivateKeyFileComponent } from './find-or-write-private-key-file-component';
import { findOrWriteDockerfileComponent } from './find-or-write-dockerfile-component';
import { ChildSpawner } from '../util/spawner/child-spawner';
import { SshSpawner } from '../util/spawner/ssh-spawner';
import { CliTerseError } from '@alwaysai/alwayscli';
import { sshDeviceFlow } from '../components/projects-and-teams/check-for-project-id';

export async function targetJsonPromptComponent(
  props: {
    targetProtocol?: TargetProtocol;
    targetHostname?: string;
    targetPath?: string;
  } = {},
) {
  const yes = false;
  const currentTargetJson = TargetJsonFile().readIfExists();
  const targetProtocol = await targetProtocolPromptComponent({
    targetProtocol:
      props.targetProtocol || (currentTargetJson && currentTargetJson.targetProtocol),
  });

  let targetHardware = 'generic';

  switch (targetProtocol) {
    case 'docker:':
      await checkForDockerComponent();
      await findOrWriteDockerfileComponent({ yes });

      const localUname = await ChildSpawner({ path: '.' }).run({
        exe: 'uname',
        args: ['-r'],
      });
      if (localUname.includes('tegra')) {
        targetHardware = 'nvidia';
        const jetpackVersion = parseFloat(
          await ChildSpawner({ path: '.' }).run({
            exe: 'sh',
            args: [
              '-c',
              'apt-cache show nvidia-jetpack | grep Version | grep -o [0-9].[0-9] | head -n1',
            ],
          }),
        );
        if (jetpackVersion > SUPPORTED_JETPACK_VERSION) {
          throw new CliTerseError(
            `alwaysAI does not support JetPack versions past ${SUPPORTED_JETPACK_VERSION}!
Please flash your device using the JetPack provided here: ${JETPACK_IMAGE_LINK}.`,
          );
        }
      }

      writeTargetJsonFileComponent({
        targetProtocol,
        dockerImageId: DOCKER_IMAGE_ID_INITIAL_VALUE,
        targetHardware,
      });
      break;

    case 'ssh+docker:':
      await findOrWriteDockerfileComponent({ yes });
      await findOrWritePrivateKeyFileComponent({ yes });
      let currentTargetHostname: string | undefined;
      let currentTargetPath: string | undefined;
      if (currentTargetJson && currentTargetJson.targetProtocol === 'ssh+docker:') {
        currentTargetHostname = currentTargetJson.targetHostname;
        currentTargetPath = currentTargetJson.targetPath;
      }
      const targetHostname = await targetHostnamePromptComponent({
        targetHostname: props.targetHostname || currentTargetHostname,
      });

      await checkForDockerComponent({ targetHostname });

      const targetPath = await targetPathPromptComponent({
        targetHostname,
        targetPath: props.targetPath || currentTargetPath,
      });

      const externalUname = await SshSpawner({ targetHostname }).run({
        exe: 'uname',
        args: ['-r'],
      });
      if (externalUname.includes('tegra')) {
        targetHardware = 'nvidia';
        const jetpackVersion = parseFloat(
          await SshSpawner({ targetHostname }).run({
            exe:
              'apt-cache show nvidia-jetpack | grep Version | grep -o [0-9].[0-9] | head -n1',
            args: [],
          }),
        );
        if (jetpackVersion > SUPPORTED_JETPACK_VERSION) {
          throw new CliTerseError(
            `alwaysAI does not support JetPack versions past ${SUPPORTED_JETPACK_VERSION}!
Please flash your device using the JetPack provided here: ${JETPACK_IMAGE_LINK}.`,
          );
        }
      }

      await writeTargetJsonFileComponent({
        targetProtocol,
        targetHostname,
        targetPath,
        dockerImageId: DOCKER_IMAGE_ID_INITIAL_VALUE,
        targetHardware,
      });

      console.log({ targetHostname });

      const hardwareId = await SshSpawner({ targetHostname }).run({
        exe:
          '( cat /var/lib/dbus/machine-id /etc/machine-id 2> /dev/null || hostname ) | head -n 1 || :',
      });

      await sshDeviceFlow(hardwareId);

      break;

    default:
      throw new Error(`Unexpected target protocol ${targetProtocol}`);
  }
}
