const fs = require('fs');
const path = require('path');
const parseStringSync = require('xml2js-parser').parseStringSync;

import { AnnotationDirName, ImageDirName } from './constants';

const DIRECTORY_STRUCTURE = `
alwaysAI retraining expects a PascalVOC datset. The zipped directory must have Pascal VOC
formatted annotation files with a "${ImageDirName}" directory containing the images corresponding to the annotation
files found in "${AnnotationDirName}"

- ${AnnotationDirName}
    - annotation xml or json files
- ${ImageDirName}
    - corresponding image files in jpeg format (.jpg or .jpeg)
`;

export function validateDataset(datasetDir: string) {
  const datasetFilesNames = fs.readdirSync(datasetDir);
  if (!datasetFilesNames.includes(AnnotationDirName)) {
    throw new Error(`${AnnotationDirName} directory not found ${DIRECTORY_STRUCTURE}`);
  }
  validateAnnotationsDirectory(path.join(datasetDir, AnnotationDirName));
  if (!datasetFilesNames.includes(ImageDirName)) {
    throw new Error(`${ImageDirName} directory not found ${DIRECTORY_STRUCTURE}`);
  }
  validateImageDirectory(path.join(datasetDir, ImageDirName));
}

function validateAnnotationsDirectory(dir: string) {
  const annotationFileNames = fs.readdirSync(dir);
  if (annotationFileNames.length === 0) {
    throw new Error(`${AnnotationDirName} directory empty ${DIRECTORY_STRUCTURE}`);
  }
  let numXmlFiles = 0;
  annotationFileNames.forEach((annotationFileName: string) => {
    if (annotationFileName.toLowerCase().endsWith('.xml')) {
      try {
        parseStringSync(fs.readFileSync(path.join(dir, annotationFileName)));
      } catch (e) {
        throw new Error(
          `Failed to parse ${annotationFileName}, The xml file might be incorrectly formatted. ${DIRECTORY_STRUCTURE} ${
            e.message
          }`,
        );
      }
      numXmlFiles += 1;
    }
  });
  if (numXmlFiles === 0) {
    throw new Error(
      `No xml files in ${AnnotationDirName} directory ${DIRECTORY_STRUCTURE}`,
    );
  }
}

function validateImageDirectory(dir: string) {
  const imageFileNames = fs.readdirSync(dir);
  if (imageFileNames.length === 0) {
    throw new Error(`${ImageDirName} directory empty ${DIRECTORY_STRUCTURE}`);
  }
  let jpgFiles = 0;
  imageFileNames.forEach((imageFileName: string) => {
    if (
      imageFileName.toLowerCase().endsWith('.jpg') ||
      imageFileName.toLowerCase().endsWith('.jpeg')
    ) {
      jpgFiles += 1;
    }
  });
  if (jpgFiles === 0) {
    throw new Error(
      `No jpg or jpeg files in ${ImageDirName} directory ${DIRECTORY_STRUCTURE}`,
    );
  }
}
